/**
 * @file framesave.c
 * @brief Saves a single frame in different formats
 * @author Florian Schmitt (0schmitt@informatik.uni-hamburg.de)
 * @date Jul-2012
**/
#include "camcapture.h"
#include <stdio.h>
#include <stdlib.h>
#include <asm/types.h>

/* globals */
int fcnt = 0;    /**< Counts the number of already recorded frames (global)*/


/**
 * Saves a given frame to a file (Y0CbY1Cr to RGB convertion is also done)
 * additionaly it appends an increasing number and .pnm to the filename
 * @param image Pointer to an image data struct
 * @param filename Filename prefix (will be ended with an id and .pnm)
 * @return 0 if success
**/
int camcapture_save_pnm_inc(const image_data_t *image, const char* filename)
{
	static int packed_value;
	static unsigned int i, width, height;
	char filegen[256];
	const void *p;
	p      = image->img;
	width  = image->width;
	height = image->height;
	sprintf(filegen, "%s-%03d.pnm", filename, fcnt++);
	FILE* fp = fopen(filegen, "w" );
	if(fp == NULL) return -1;

	/* Write PNM header */
	fprintf(fp, "P6\n");
	fprintf(fp, "# YUV422 frame -> RGB \n");
	fprintf(fp, "%d %d\n", width, height);

	int Y0, Y1, Cb, Cr;            /* gamma pre-corrected input [0;255] */
	int ER0,ER1,EG0,EG1,EB0,EB1;   /* output [0;255] */
	double r0,r1,g0,g1,b0,b1;      /* temporaries */
	double y0,y1, pb, pr;

	/* Max val */
	fprintf(fp, "255\n");

	/* loop thrugh all points */
	i = 0;
	while(i < (width * height/2))
	{
		packed_value = *((int*)p+i);

		Y0 = (char)(packed_value & 0xFF);
		Cb = (char)((packed_value >> 8) & 0xFF);
		Y1 = (char)((packed_value >> 16) & 0xFF);
		Cr = (char)((packed_value >> 24) & 0xFF);

		/* Strip sign values after shift (i.e. unsigned shift) */
		Y0 = Y0 & 0xFF;
		Cb = Cb & 0xFF;
		Y1 = Y1 & 0xFF;
		Cr = Cr & 0xFF;

		y0 = (255 / 219.0) * (Y0 - 16);
		y1 = (255 / 219.0) * (Y1 - 16);
		pb = (255 / 224.0) * (Cb - 128);
		pr = (255 / 224.0) * (Cr - 128);

		/* Generate first pixel */
		r0 = 1.0 * y0 + 0     * pb + 1.402 * pr;
		g0 = 1.0 * y0 - 0.344 * pb - 0.714 * pr;
		b0 = 1.0 * y0 + 1.772 * pb + 0     * pr;

		/* Generate next pixel - must reuse pb & pr as 4:2:2 */
		r1 = 1.0 * y1 + 0     * pb + 1.402 * pr;
		g1 = 1.0 * y1 - 0.344 * pb - 0.714 * pr;
		b1 = 1.0 * y1 + 1.772 * pb + 0     * pr;

		ER0 = clamp (r0);
		ER1 = clamp (r1);
		EG0 = clamp (g0);
		EG1 = clamp (g1);
		EB0 = clamp (b0);
		EB1 = clamp (b1);

		fprintf(fp, "%c%c%c%c%c%c", ER0,EG0,EB0,ER1,EG1,EB1);  /* Output two pixels */
		i++;
	}

	/* work done */
	fclose(fp);
	return 0;
}



/**
 * Saves a given frame to a file (Y0CbY1Cr to RGB convertion is also done)
 * appends nothing to the filename
 * @param image Pointer to an image data struct
 * @param filename Filename prefix (will be ended with an id and .pnm)
 * @return 0 if success
**/
int camcapture_save_pnm(const image_data_t *image, const char* filename)
{
	static int packed_value;
	static unsigned int i, width, height;
	const void *p;
	p      = image->img;
	width  = image->width;
	height = image->height;
	FILE* fp = fopen(filename, "w" );

	/* Write PNM header */
	fprintf(fp, "P6\n");
	fprintf(fp, "# YUV422 frame -> RGB \n");
	fprintf(fp, "%d %d\n", width, height);

	int Y0, Y1, Cb, Cr;            /* gamma pre-corrected input [0;255] */
	int ER0,ER1,EG0,EG1,EB0,EB1;   /* output [0;255] */
	double r0,r1,g0,g1,b0,b1;      /* temporaries */
	double y0,y1, pb, pr;

	/* Max val */
	fprintf(fp, "255\n");

	/* loop thrugh all points */
	i = 0;
	while(i < (width * height/2))
	{
		packed_value = *((int*)p+i);

		Y0 = (char)(packed_value & 0xFF);
		Cb = (char)((packed_value >> 8) & 0xFF);
		Y1 = (char)((packed_value >> 16) & 0xFF);
		Cr = (char)((packed_value >> 24) & 0xFF);

		/* Strip sign values after shift (i.e. unsigned shift) */
		Y0 = Y0 & 0xFF;
		Cb = Cb & 0xFF;
		Y1 = Y1 & 0xFF;
		Cr = Cr & 0xFF;

		y0 = (255 / 219.0) * (Y0 - 16);
		y1 = (255 / 219.0) * (Y1 - 16);
		pb = (255 / 224.0) * (Cb - 128);
		pr = (255 / 224.0) * (Cr - 128);

		/* Generate first pixel */
		r0 = 1.0 * y0 + 0     * pb + 1.402 * pr;
		g0 = 1.0 * y0 - 0.344 * pb - 0.714 * pr;
		b0 = 1.0 * y0 + 1.772 * pb + 0     * pr;

		/* Generate next pixel - must reuse pb & pr as 4:2:2 */
		r1 = 1.0 * y1 + 0     * pb + 1.402 * pr;
		g1 = 1.0 * y1 - 0.344 * pb - 0.714 * pr;
		b1 = 1.0 * y1 + 1.772 * pb + 0     * pr;

		ER0 = clamp (r0);
		ER1 = clamp (r1);
		EG0 = clamp (g0);
		EG1 = clamp (g1);
		EB0 = clamp (b0);
		EB1 = clamp (b1);

		fprintf(fp, "%c%c%c%c%c%c", ER0,EG0,EB0,ER1,EG1,EB1);  /* Output two pixels */
		i++;
	}

	/* work done */
	fclose(fp);
	return 0;
}
