/**
 * @file lists.c
 * @brief Provides functions for lists
 * @author Florian Schmitt (0schmitt@informatik.uni-hamburg.de)
 * @date Aug-2012
**/
#include "camcapture.h"
#include <math.h>
#include <stdlib.h>

/** @cond */
#define PI 3.14159265
/** @endcond */


/**
 * Creates a linked list of image point
 * @return the new start node
**/
image_point_list_t *camcapture_create_point_list()
{
	/* allocate memory */
	image_point_list_t *obj = malloc(sizeof(image_point_list_t));
	obj->next = 0;
	obj->data = 0;
	return obj;
}

/**
 * Clean a whole list up
 * @param obj The image point list object to clean up (pointer)
**/
void camcapture_destroy_point_list(image_point_list_t *obj)
{
	image_point_list_t *next = 0;
	/* until there is no follower */
	while(obj != 0)
	{
		next = obj->next;
		if(obj->data != 0) free(obj->data);
		free(obj);
		obj = next;
	}
}

/**
 * Append a point to a pointlist (inserts at the end). 
 * NOTICE: This is pretty unefficient! Try to use camcapture_add_point_list() !!!
 * @param list The image point list reference
 * @param point The image point to append
 * @return the appended objects adress
**/
image_point_list_t *camcapture_append_point_list(image_point_list_t *list, image_point_t *point)
{
	if(list == 0) return 0;
	/* until there is no follower */
	while(list->next != 0)
	{
		list = list->next;
	}
	/* create new node */
	list->next = camcapture_create_point_list();
	list = list->next;
	/* set data */
	list->data = point;
	/* info: it's not needed to set appended follower to 0, 
	 * because the create_point_list-Function does already. */
	return list;
}


/**
 * Add a point to a pointlist (inserts at the beginning)
 * @param list The image point list reference
 * @param point The image point to add
 * @return the appended objects adress
**/
image_point_list_t *camcapture_add_point_list(image_point_list_t *list, image_point_t *point)
{
	if(list == 0) return 0;
	image_point_list_t *follower = list->next;
	/* create new node */
	list->next = camcapture_create_point_list();
	list = list->next;
	/* set data */
	list->data = point;
	/* set follower */
	list->next = follower;
	return list;
}


/**
 * Get next object
 * @param list The image point list reference
 * @return The next point list element
**/
image_point_list_t *camcapture_next_point_list(image_point_list_t *list)
{
	if(list == 0) return 0;
	return list->next;
}


/**
 * Count objects in a list
 * @param list The image point list reference
 * @return The number of elements in the list
**/
int camcapture_point_list_count(image_point_list_t *list)
{
	if(list == 0) return -1;
	int ret = 0;
	while( (list = list->next) ) ret++;
	return ret;
}

/**
 * Calculates the median
 * @param list The image point list reference
 * @return The median point as an image_point_t
**/
image_point_t camcapture_point_list_median(image_point_list_t *list)
{
	int cnt = 0;
	image_point_t ret;
	ret.x = 0; ret.y = 0;
	/* ignore first none */
	if(list == 0) return ret;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			ret.x += list->data->x;
			ret.y += list->data->y;
		}
		cnt++;
	}
	if(cnt > 0)
	{
		ret.x = ret.x / cnt;
		ret.y = ret.y / cnt;
	}
	return ret;
}


/**
 * Kills points who variate more then a certain distance from a given point
 * @param list The image point list reference
 * @param point Center of the calculation
 * @param maxdist Radius around the point
 * @return Number of eliminated nodes
**/
int camcapture_point_kill_overshoots(image_point_list_t *list, image_point_t point, int maxdist)
{
	float dist = 0.0f;
	int cnt = 0;
	int tmp1, tmp2;
	image_point_list_t *previous = list;
	/* ignore first none */
	if(list == 0) return 0;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			/* calculate distance */
			tmp1  = list->data->x - point.x;
			tmp1 *= tmp1;
			tmp2  = list->data->y - point.y;
			tmp2 *= tmp2;
			dist  = sqrt(tmp1 + tmp2);
			/* on distance overshoot: kill node */
			if( (int)dist > maxdist )
			{
				previous->next = list->next;
				free(list->data);
				free(list);
				list = previous;
				cnt++;
			}
		}
		/* get previous element before next loop */
		previous = list;
	}
	return cnt;
}






/**
 * This is a helper function for camcapture_point_kill_nonobjs().
 * It checks a point in an image if its value is above 200.
 * @param x position in image
 * @param y position in image
 * @param image Image data structure
 * @return 1 if point has a higher Y0 value than 200, otherwise 0
**/
int kill_nonobjs_isPointWhite(int x, int y, image_data_t *image)
{
	int *data = (int *)image->img;
	/* check if point is inside the image */
	if( x < 0 )             return 0;
	if( x > image->width )  return 0;
	if( y < 0 )             return 0;
	if( y > image->height ) return 0;
	/* check if color is white */
	data = data + (x + y*image->width)/2;
	return ((*data) & 0xFF) > 200;
}




/**
 * Kills points who have no object in their background (white filtered picture). 
 * This goes thru all points and kills them if they dont have white area at minumum 
 * 4 of 8 points around them. 
 * @param list The image point list reference
 * @param image Pointer to the image object data structure
 * @param distance The distance of the color check around a point
 * @return Number of eliminated nodes
**/
int camcapture_point_kill_nonobjs(image_point_list_t *list, image_data_t *image, int distance)
{
	int cnt = 0;
	image_point_list_t *previous = list;
	/* ignore first none */
	if(list == 0) return 0;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			int checked = 0;
			
			if(kill_nonobjs_isPointWhite(list->data->x + distance, list->data->y, image)) checked++;
			if(kill_nonobjs_isPointWhite(list->data->x - distance, list->data->y, image)) checked++;
			if(kill_nonobjs_isPointWhite(list->data->x, list->data->y + distance, image)) checked++;
			if(kill_nonobjs_isPointWhite(list->data->x, list->data->y - distance, image)) checked++;
			
			if(kill_nonobjs_isPointWhite(list->data->x - distance, list->data->y - distance, image)) checked++;
			if(kill_nonobjs_isPointWhite(list->data->x + distance, list->data->y + distance, image)) checked++;
			if(kill_nonobjs_isPointWhite(list->data->x + distance, list->data->y - distance, image)) checked++;
			if(kill_nonobjs_isPointWhite(list->data->x - distance, list->data->y + distance, image)) checked++;
			
			/* kill node */
			if( checked < 4 )
			{
				previous->next = list->next;
				free(list->data);
				free(list);
				list = previous;
				cnt++;
			}
		}
		/* get previous element before next loop */
		previous = list;
	}
	return cnt;
}




/**
 * This is a helper function for camcapture_point_kill_nonobjs_col().
 * It checks if a point in an image is of the given color FCb and FCr.
 * @param x position in image
 * @param y position in image
 * @param sens Sensibility for the camcapture_filter_color_single-function
 * @param FCb Color value (axis Cb)
 * @param FCr Color value (axis Cr)
 * @param image Image data structure
 * @return 1 if point is of the given color FCb and FCr
**/
int kill_nonobjs_isPointColor(int x, int y, int sens, int FCb, int FCr, image_data_t *image)
{
	int *data = (int *)image->img;
	/* check if point is inside the image */
	if( x < 0 )             return 0;
	if( x > image->width )  return 0;
	if( y < 0 )             return 0;
	if( y > image->height ) return 0;
	/* check if color is white */
	data = data + (x + y*image->width)/2;
	/* check */
	return camcapture_filter_color_single(data, FCb, FCr, sens);
}



/**
 * Kills points who have no object in their background (white filtered picture). 
 * This goes thru all points and kills them if they dont have white area at minumum 
 * 4 of 8 points around them. 
 * @param list The image point list reference
 * @param image Pointer to the image object data structure
 * @param distance The distance of the color check around a point
 * @return Number of eliminated nodes
**/
int camcapture_point_kill_nonobjs_col(image_point_list_t *list, image_data_t *image, int distance, int sens, color_rgb_t *color)
{
	int FCb, FCr;
	int cnt = 0;
	rgb_to_ycbcr(color, 0, &FCb, &FCr);
	image_point_list_t *previous = list;
	/* ignore first none */
	if(list == 0) return 0;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			int checked = 0;
			
			if(kill_nonobjs_isPointColor(list->data->x + distance, list->data->y, sens, FCb, FCr, image)) checked++;
			if(kill_nonobjs_isPointColor(list->data->x - distance, list->data->y, sens, FCb, FCr, image)) checked++;
			if(kill_nonobjs_isPointColor(list->data->x, list->data->y + distance, sens, FCb, FCr, image)) checked++;
			if(kill_nonobjs_isPointColor(list->data->x, list->data->y - distance, sens, FCb, FCr, image)) checked++;
			
			if(kill_nonobjs_isPointColor(list->data->x - distance, list->data->y - distance, sens, FCb, FCr, image)) checked++;
			if(kill_nonobjs_isPointColor(list->data->x + distance, list->data->y + distance, sens, FCb, FCr, image)) checked++;
			if(kill_nonobjs_isPointColor(list->data->x + distance, list->data->y - distance, sens, FCb, FCr, image)) checked++;
			if(kill_nonobjs_isPointColor(list->data->x - distance, list->data->y + distance, sens, FCb, FCr, image)) checked++;
			
			/* kill node */
			if( checked < 4 )
			{
				previous->next = list->next;
				free(list->data);
				free(list);
				list = previous;
				cnt++;
			}
		}
		/* get previous element before next loop */
		previous = list;
	}
	return cnt;
}


/**
 * Calculates the mean value of distances to a given median.
 * In a circle this would be the radius.
 * @param list The image point list reference
 * @param median The median point
 * @return Mean value of the distances (if object is a circle: the radius)
**/
int camcapture_point_list_median_delta(image_point_list_t *list, image_point_t median)
{
	float ret = 0.0f;
	int cnt = 0;
	int tmp1, tmp2;
	/* ignore first node */
	if(list == 0) return (int) ret;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			tmp1  = list->data->x - median.x;
			tmp1 *= tmp1;
			tmp2  = list->data->y - median.y;
			tmp2 *= tmp2;
			ret += sqrt(tmp1 + tmp2);
		}
		cnt++;
	}
	if((ret == 0.0f) || (cnt == 0)) return 0;
	return (int) (ret/cnt);
}


/**
 * Calculates maximum distance to a given median.
 * In a circle this would be the (maximum) radius.
 * @param list The image point list reference
 * @param median The median point
 * @return maximum distance to median (if object is a circle: the radius)
**/
int camcapture_point_list_max_delta(image_point_list_t *list, image_point_t median)
{
	float ret = 0.0f;
	int tmp1, tmp2;
	float tmp3;
	/* ignore first node */
	if(list == 0) return (int) ret;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			tmp1  = list->data->x - median.x;
			tmp1 *= tmp1;
			tmp2  = list->data->y - median.y;
			tmp2 *= tmp2;
			tmp3  = sqrt(tmp1 + tmp2);
			if(tmp3 > ret) ret = tmp3;
		}
	}
	return (int) ret;
}


/**
 * Calculates maximum x point
 * @param list The image point list reference
 * @return Maximum x points value
**/
int camcapture_point_list_max_x(image_point_list_t *list)
{
	int max = 0;
	/* ignore first node */
	if(list == 0) return 0;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			if(list->data->x > max) max = list->data->x;
		}
	}
	return max;
}


/**
 * Calculates minimum x point
 * @param list The image point list reference
 * @return Minumum x points value
**/
int camcapture_point_list_min_x(image_point_list_t *list)
{
	int min = 9999;
	/* ignore first node */
	if(list == 0) return 0;
	if(list->next == 0) return 0;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			if(list->data->x < min) min = list->data->x;
		}
	}
	return min;
}





/**
 * Calculates maximum y point
 * @param list The image point list reference
 * @return Maximum y points value
**/
int camcapture_point_list_max_y(image_point_list_t *list)
{
	int max = 0;
	/* ignore first node */
	if(list == 0) return 0;
	/* loop list */
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			if(list->data->y > max) max = list->data->y;
		}
	}
	return max;
}


/**
 * Calculates minimum y point
 * @param list The image point list reference
 * @return Minumum y points value
**/
int camcapture_point_list_min_y(image_point_list_t *list)
{
	int min = 9999;
	/* ignore first node */
	if(list == 0) return 0;
	/* loop list */
	if(list->next == 0) return 0;
	while( (list = list->next) != 0 )
	{
		if(list->data != 0)
		{
			if(list->data->y < min) min = list->data->y;
		}
	}
	return min;
}


